#include <png/png.h>

// --- input
#define KEY_MENU  0
#define KEY_LEFT  1
#define KEY_RIGHT 2
#define KEY_UP    3
#define KEY_DOWN  4
#define KEY_A     5
#define KEY_B     6
bool KeyPressed(int key)
{
	switch(key)
	{
	case KEY_MENU:  return (GetAsyncKeyState(VK_ESCAPE)&KF_UP);
	case KEY_LEFT:  return (GetAsyncKeyState(VK_LEFT)&KF_UP);
	case KEY_RIGHT: return (GetAsyncKeyState(VK_RIGHT)&KF_UP);
	case KEY_UP:    return (GetAsyncKeyState(VK_UP)&KF_UP);
	case KEY_DOWN:  return (GetAsyncKeyState(VK_DOWN)&KF_UP);
	case KEY_A:     return (GetAsyncKeyState('Z')&KF_UP);
	case KEY_B:     return (GetAsyncKeyState('X')&KF_UP);
	}
	return false;
}
// ---

// --- memory handler
void* mh_pointers[1024];
int mh_num=0;
void* mh_malloc(int numbytes)
{
	void* ptr=malloc(numbytes);
	mh_pointers[mh_num++]=ptr;
	return ptr;
}
void mh_free()
{
	for(int i=0;i<mh_num;i++) free(mh_pointers[i]);
	mh_num=0;
}
// ---

// --- error handling
void Quit()
{
	mh_free();
	ptc_close();
	ExitProcess(0);
}
void check(bool ok)
{
	if(!ok) Quit();
}
// ---

// --- misc
int imax(int a, int b) { return a>b?a:b; }
int imin(int a, int b) { return a<b?a:b; }
// ---

// --- graphics
DWORD screen[320*240];
struct Image { int width; int height; DWORD* data; };
void DrawImage(Image& img, int sx, int sy, bool white)
{
	DWORD wc=0x000000;
	if(white) wc=0xFFFFFF;
	for(int y=sy;y<sy+img.height;y++)
	{
		if(y<0) continue; if(y>=240) break; // y clip
		DWORD* sp=img.data+(y-sy)*img.width; // source pointer (image)
		DWORD* dp=screen+y*320+sx; // destination pointer (screen)
		for(int x=sx;x<sx+img.width;x++,sp++,dp++)
		{
			if(x<0) continue; if(x>=320) break; // x clip
			DWORD pixel=*sp; // read pixel
			if(pixel!=0x604020) *dp=pixel|wc; // check color key and write opaque pixels
		}
	}
}
DWORD fxbuffer[320*64];
Image font, gradient;
void DrawFXChar(int sx, int sy, int mode, char chi)
{
	if(chi>='-' && chi<='9') chi-='-';
	else if(chi>='A' && chi<='Z') chi=chi-'A'+13;
	else if(chi=='?') chi=39;
	else if(chi=='!') chi=2;
	else return;
	for(int y=sy;y<sy+16;y++)
	{
		if(y<0) continue; if(y>=64) break;
		DWORD* sp=font.data+(y-sy)*font.width+chi*16;
		DWORD* dp=fxbuffer+y*320+sx;
		for(int x=sx;x<sx+16;x++,sp++,dp++)
		{
			if(x<0) continue; if(x>=320) break;
			DWORD pixel=*sp;
//			if(pixel==0x604020) continue;
//			if(pixel==0xFFFFFF) *dp=pixel;
			*dp=pixel;
//			if(pixel!=0x604020) *dp=pixel;
		}
	}
}
void DrawString(char* string, int slen, int offset)
{
	for(int i=offset/16-1;i<offset/16+320/16+1;i++)
	{
		if(i<0) continue;
			DrawFXChar(i*16-offset, 24, 0, string[i%slen]);
	}
}
void ClearFXBuffer()
{
	for(int i=0;i<320*64;i++)
		fxbuffer[i]=0x604020;
}
float sine_phase=0.0f;
float sine_pamp=0.0f;
void Overlay()
{
	sine_phase-=0.019f;
	sine_pamp+=0.015f;
	int i=0;
	for(int y=0;y<64;y++)
	{
		DWORD gradc=gradient.data[(y+(int)(sine_pamp*10))%64*4];
		for(int x=0;x<320;x++,i++)
		{
			int sy=y+(int)(sin((float)x*0.05f+sine_phase)*(6.0f+cos(sine_pamp+(float)x*0.02f)*4.0f));
			if(sy<0 || sy>=64) continue;
			DWORD c=fxbuffer[sy*320+x];
			if(c==0xFFFFFF) screen[i]=c;
			else if(c!=0x604020) screen[i]=gradc;
		}
	}
}
void DrawBar(int sx, int sy, int w, int h, DWORD color)
{
	for(int y=sy;y<sy+h;y++)
	{
		if(y<0) continue; if(y>=240) break;
		DWORD* dp=screen+y*320+sx;
		for(int x=sx;x<sx+w;x++,dp++)
		{
			if(x<0) { continue; } if(x>=320) break;
			*dp=color;
		}
	}
}
void UpdateScreen(bool filter) // stretch 320x240 to 640x480, with optional filtering
{
	DWORD* dp=ptc_screen;
	DWORD* sp=screen;
	if(filter)
		for(int y=0;y<240;y++)
		{
			for(int x=0;x<319;x++) // even row
			{ *dp++=*sp; *dp++=(((*(sp+1))>>1)&0x7F7F7F)+(((*sp++)>>1)&0x7F7F7F); }
			*dp++=*sp; *dp++=*sp++; // special case for last pixel to avoid artifacts
			sp-=320; // step back one row in source buffer
			for(int x=0;x<319;x++) // odd row (half brightness)
			{ *dp++=(*sp>>1)&0x7F7F7F; *dp++=(((*(sp+1))>>2)&0x3F3F3F)+(((*sp++)>>2)&0x3F3F3F); }
			*dp++=(*sp>>1)&0x7F7F7F; *dp++=((*sp++)>>1)&0x7F7F7F; // special case for last pixel
		}
	else
		for(int y=0;y<240;y++)
		{
			for(int x=0;x<320;x++) // even row
			{ *dp++=*sp; *dp++=*sp++; }
			sp-=320; // step back one row in source buffer
			for(int x=0;x<320;x++) // odd row
			{ *dp++=*sp; *dp++=*sp++; }
		}
}
bool LoadPNG(Image& img, char* filename)
{
	FILE *fp=fopen(filename, "rb"); if(!fp) return false; // open file
	png_structp png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL); if(!png_ptr) return false;
	png_infop info_ptr=png_create_info_struct(png_ptr);
	if(!info_ptr) { png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL); return false; }
	png_init_io(png_ptr, fp); png_read_info(png_ptr, info_ptr); png_read_update_info(png_ptr, info_ptr);
	img.width=info_ptr->width; img.height=info_ptr->height; int channels=info_ptr->channels; // get image info
	BYTE *rgb_image=(BYTE*)malloc(img.width*img.height*channels); // temporary buffer
	img.data=(DWORD*)mh_malloc(img.width*img.height*sizeof(DWORD)); // target buffer in image
	png_byte **row_pointers=(png_byte**)malloc(img.height*sizeof(png_byte*));
	for(int i=0;i<img.height;i++) row_pointers[i]=(png_byte*)(rgb_image+i*img.width*channels);
	png_set_rows(png_ptr, info_ptr, (png_bytepp)(&row_pointers));
	png_read_image(png_ptr, row_pointers); png_read_end(png_ptr, NULL); // read png image
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL); fclose(fp); free(row_pointers);
	DWORD* dp=img.data; int si=0; // destination pointer and source pixel index
	for(int i=0;i<img.width*img.height;i++) // copy 24-bit pixel data to 32-bit image buffer
	{
		DWORD pixel=(rgb_image[si+0]<<16)|(rgb_image[si+1]<<8)|rgb_image[si+2];
		*dp++=pixel; si+=channels;
	}
	free(rgb_image); return true;
}
// ---

// --- isometry
struct IsoQEntry { Image* img; int x, y, depth; bool white; };
IsoQEntry isoqueue[1024];
int isonum=0;
int IsoX(int x, int y, int z) { return 160+x-y; }
int IsoY(int x, int y, int z) { int sy=y/2+x/2-z; if((x&1)&&(y&1)) return sy+1; else return sy;}
void DrawIso(Image& img, int x, int y, int z, bool above, bool white)
{
	IsoQEntry& iq=isoqueue[isonum];
	iq.x=160+x-y-scrollx; iq.y=y/2+x/2-z-scrolly;
	if((x&1) && (y&1)) iq.y++; // remove vertical jitter
	if(iq.x>=320 || iq.y>=240 || iq.x+img.width<=0 || iq.y+img.height<=0) return; // cull off-screen objects
	iq.img=&img; iq.depth=y/2+x/2+z;
	if(above) iq.depth+=16;
	iq.white=white;
	isonum++;
}
void FlushIso()
{
	// bubble sort all queued objects based on depth
	for(bool sorted=false;!sorted;)
	{
		sorted=true;
		for(int i=1;i<isonum;i++)
			if(isoqueue[i-1].depth>isoqueue[i].depth)
			{ IsoQEntry temp=isoqueue[i]; isoqueue[i]=isoqueue[i-1]; isoqueue[i-1]=temp; sorted=false; }
	}
	// draw them
	IsoQEntry* iq;
	for(int i=0;iq=&isoqueue[i],i<isonum;i++)
		DrawImage(*iq->img, iq->x, iq->y, iq->white);
	isonum=0;
}
// ---

